/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is ForumZilla 0.1.1.
 *
 * The Initial Developer of the Original Code is Zapogee, Inc.
 * Portions created by Zapogee, Inc. are Copyright (C) 2001
 * Zapogee, Inc.  All Rights Reserved.
 *
 * Contributor(s):
 *   Myk Melez, <myk@melez.com>, original author
 */

// Feed service constructor.  Note that we want the feed service to be a singleton,
// but it's possible to instantiate multiple instances at the moment since it isn't
// a real XPCOM component.
function FeedService() {
    this.itemStorageInitiated = new Date();
}

// This should only ever be called once per application session when the mail window
// first loads.
FeedService.prototype.initFeedPolling = function() {
    // Flag to indicate whether or not we're in the process of storing an item.
    // Used by storeItem() to determine when to start storing the next item.
    this.itemStorageUnderway = false;

    // Check for new stories.
    feedService.downloadFeeds();

    // Save data sources to disk periodically in case host application crashes.
    window.setInterval("feedService.flushDataSources();", 2 * 60 * 1000);

    // Periodically check for new stories.
    window.setInterval("feedService.downloadFeeds();", 60 * 1000);

    // Periodically store items in the storage queue into their message folders.
    //window.setInterval("feedService.storeItems();", 5 * 1000);
    window.setInterval("feedService.storeItem();", 100);
}
    
FeedService.prototype.__defineGetter__('pollingInterval', function() {
    return nsPreferences.getIntPref("forumzilla.polling.interval", DEFAULT_POLLING_INTERVAL);
});

FeedService.prototype.__defineGetter__('storageMode', function() {
    return nsPreferences.getIntPref("forumzilla.storage.mode", DEFAULT_STORAGE_MODE);
});

FeedService.prototype.__defineGetter__('storageFolder', function() {
    return nsPreferences.copyUnicharPref("forumzilla.storage.folder", DEFAULT_STORAGE_FOLDER);
});

// Download a feed.
FeedService.prototype.downloadFeed = function(resource) {
    this.getFeed(resource).download();
}

// Cache feeds so we don't keep creating them over and over and can hang
// XMLHttpRequest objects off them that they can retrieve after the request
// is complete.
FeedService.feeds = new Object();

FeedService.prototype.getFeed = function(id) {
    var url;
    try {
        // If the identifier is a resource, extract its URL.
        var resource = id.QueryInterface(Components.interfaces.nsIRDFResource);
        url = this.subscriptionsDS.GetTarget(resource, DC_IDENTIFIER, true);
        url = url.QueryInterface(Components.interfaces.nsIRDFLiteral);
        url = url.Value;
    }
    catch(e) {
        // The identifier is not a resource, so it should be a URL.
        url = id;
    }
    var feed = FeedService.feeds[url];
    if (!feed) {
        debug(url + " requested from feed service, but not in cache; creating anew");
        feed = new Feed(resource);
        FeedService.feeds[feed.url] = feed;
    }
    else
        debug(url + " requested from feed service, and found in cache");
    return feed;
}

// Download feeds and check for new messages.
FeedService.prototype.downloadFeeds = function() {
    debug("polling feeds");

    var ds = this.subscriptionsDS;
    var feeds = this.subscriptionsList.GetElements();
  
    while(feeds.hasMoreElements()) {
        var resource = feeds.getNext();
        var feed = this.getFeed(resource);

        // Don't poll the feed if it has been disabled.
        if (!feed.enabled) {
            debug(feed.url + " not downloaded; disabled");
            continue;
        }

        // Don't poll the feed if it isn't time yet.
        var now = new Date().getTime();
        var then = feed.lastPolled ? feed.lastPolled.getTime() : 0;
        var elapsed = (now - then)/1000/60;
        if (elapsed < feed.interval) {
            debug(feed.url + " not downloaded; " + elapsed.toPrecision(2) + "/" + feed.interval);
            continue;
        }

        feed.download();
    }
}


// Items to store (i.e. items that have been downloaded but haven't been stored
// into a mail folder yet).  Processed by storeItems().
FeedService.itemsToStore = new Array();

FeedService.prototype.queueItem = function(item) {
    FeedService.itemsToStore.push(item);
}

// NOTE: storeItems() isn't used at the moment because of a bug in the copy
// service that causes it not to call the copy listener's OnStopCopy() method
// sometimes when copying multiple messages to an IMAP folder.  Instead, we
// copy messages one at a time by setting a flag to indicate when a message
// is being copied and calling storeItem() repeatedly to copy the next message
// if the flag indicates the previous copy has completed.

/*
// Stores the items in the queue.
FeedService.prototype.storeItems = function() {
    item = FeedService.itemsToStore.shift();
    while (item) {
        item.store();
        item = FeedService.itemsToStore.shift();
    }
}

*/


// Stores the next item in the queue.
// We initiate storage of the next item if the last item has finished
// being stored or if 60 seconds have passed since it was initiated to make
// sure we always continue storing items even if some item storage gets stuck
// and never completes (and thus never clears the "itemStorageUnderway" flag).
FeedService.prototype.storeItem = function() {
    var elapsed = new Date().getTime() - this.itemStorageInitiated.getTime();
    if (!this.itemStorageUnderway || (elapsed > 60*1000)) {
        if (elapsed > 60*1000)
            debug("FS.storeItem: " + elapsed + " elapsed since last storage initiated");
        this.itemStorageUnderway = true;
        this.itemStorageInitiated = new Date();
        item = FeedService.itemsToStore.shift();
        if (item) {
            try {
                item.store();
            }
            catch(e) {
                this.itemStorageUnderway = false;
                debug('FS.storeItem: error attempting to store item: ' + e);
            }
        }
        else
            this.itemStorageUnderway = false;
    }
}

// Feeds to parse.
FeedService.feedsToParse = new Object();

FeedService.prototype.flushDataSources = function() {
    var ds;

    try {
        ds = this.subscriptionsDS;
        ds = ds.QueryInterface(Components.interfaces.nsIRDFRemoteDataSource);
        if (ds)
            ds.Flush();
    }
    catch(e) {
        debug("couldn't flush subscriptions data source: " + e);
    }

    try {
        ds = this.itemsDS;
        ds = ds.QueryInterface(Components.interfaces.nsIRDFRemoteDataSource);
        if (ds)
            ds.Flush();
    }
    catch(e) {
        debug("couldn't flush items data source: " + e);
    }
}

FeedService.prototype.createFolder = function(uri) {
    debug("trying to create folder " + uri);
    var folder;
    try {
        folder = rdf.GetResource(uri).QueryInterface(Components.interfaces.nsIMsgFolder);
        var urlListener = {
            OnStartRunningURL: function(url) {},
            OnStopRunningURL: function(url, exitCode) {}
        };
        folder.createStorageIfMissing(urlListener);
    }
    catch(e) {
        debug("error creating folder: " + e);
    }
    return folder;
}

FeedService.prototype.createDefaultFolder = function(title) {
    var uri;

    if (feedService.storageMode == FOLDER_STORAGE_MODE)
        uri = feedService.storageFolder;

    else if (feedService.storageMode == SUBFOLDERS_STORAGE_MODE) {
        var parent = GetMsgFolderFromUri(feedService.storageFolder, true);
        if (!parent.containsChildNamed(title))
            parent.createSubfolder(title, this.msgWindow);

        var child = parent.getChildNamed(title);
        child.QueryInterface(Components.interfaces.nsIMsgFolder);

        uri = child.URI;
    }

    return uri;
}

FeedService.prototype.__defineGetter__('msgWindow',
function() {
    var rv;
    if (window.msgWindow)
        rv = msgWindow;
    else if (window.opener && window.opener.msgWindow)
        rv = window.opener.msgWindow;
    else if (window.opener.opener && window.opener.opener.msgWindow)
        rv = window.opener.opener.msgWindow;
    return rv;
});

FeedService.prototype.__defineGetter__('subscriptionsDS',
function() {
    if (this._subscriptionsDS)
        return this._subscriptionsDS;

    var url = fileHandler.getURLSpecFromFile(this.subscriptionsFile);

    this._subscriptionsDS = rdf.GetDataSource(url);
    if (!this._subscriptionsDS)
        throw("can't get subscriptions data source");

    // Note that it this point the datasource may not be loaded yet.
    // You have to QueryInterface it to nsIRDFRemoteDataSource and check
    // its "loaded" property to be sure.  You can also attach an observer
    // which will get notified when the load is complete.

    return this._subscriptionsDS;
});

FeedService.prototype.__defineGetter__('subscriptionsFile',
function() {
    // Get the user's profile directory.
    var profileDir =
        appDirectoryService.get("ProfD", Components.interfaces.nsIFile);
    if (!profileDir)
        throw ("couldn't get the user's profile directory");
  
    // Get the user's subscriptions file.
    var file = profileDir.clone();
    file.append("feeds.rdf");

    // If the file doesn't exist, create it.
    if (!file.exists())
        this.createSubscriptionsFile(file);
  
    return file;
});

FeedService.prototype.createSubscriptionsFile = function(file) {
    var fileout = new LocalFile(file, MODE_WRONLY | MODE_CREATE);
    fileout.write('\
<?xml version="1.0"?>\n\
<RDF:RDF xmlns:dc="http://purl.org/dc/elements/1.1/"\n\
         xmlns:fz="' + FZ_NS + '"\n\
         xmlns:RDF="http://www.w3.org/1999/02/22-rdf-syntax-ns#">\n\
  <RDF:Description about="' + FZ_ROOT.Value + '">\n\
    <fz:feeds>\n\
      <RDF:Seq>\n\
      </RDF:Seq>\n\
    </fz:feeds>\n\
  </RDF:Description>\n\
</RDF:RDF>\n\
');
    fileout.close();
}

FeedService.prototype.__defineGetter__('itemsDS',
function() {
    if (this._itemsDS)
        return this._itemsDS;

    var url = fileHandler.getURLSpecFromFile(this.itemsFile);

    this._itemsDS = rdf.GetDataSource(url);
    if (!this._itemsDS)
        throw("can't get items data source");

    // Note that it this point the datasource may not be loaded yet.
    // You have to QueryInterface it to nsIRDFRemoteDataSource and check
    // its "loaded" property to be sure.  You can also attach an observer
    // which will get notified when the load is complete.

    return this._itemsDS;
});

FeedService.prototype.__defineGetter__('itemsFile',
function() {
    // Get the user's profile directory.
    var profileDir =
        appDirectoryService.get("ProfD", Components.interfaces.nsIFile);
    if ( !profileDir )
        throw ("couldn't get the user's profile directory");
  
    // Get the user's subscriptions file.
    var file = profileDir.clone();
    file.append("feeditems.rdf");
  
    // If the file doesn't exist, create it.
    if (!file.exists()) {
        var fileout = new LocalFile(file, MODE_WRONLY | MODE_CREATE);
        fileout.write('\
<?xml version="1.0"?>\n\
<RDF:RDF xmlns:dc="http://purl.org/dc/elements/1.1/"\n\
         xmlns:fz="' + FZ_NS + '"\n\
         xmlns:RDF="http://www.w3.org/1999/02/22-rdf-syntax-ns#">\n\
</RDF:RDF>\n\
');
        fileout.close();
    }
    return file;
});

FeedService.prototype.__defineGetter__('subscriptionsList',
function() {
    var list = this.subscriptionsDS.GetTarget(FZ_ROOT, FZ_FEEDS, true);
    //list = feeds.QueryInterface(Components.interfaces.nsIRDFContainer);
    list = list.QueryInterface(Components.interfaces.nsIRDFResource);
    list = rdfcontainer.MakeSeq(this.subscriptionsDS, list);
    return list;
});
